import abc
from typing import Any

from .comp_protocol import ComponentProtocol
from .func_parser import FunctionParser

import pytest


class ProtocolCompSuite(abc.ABC):
    @pytest.fixture
    @abc.abstractmethod
    def comp_fixture(self, *args: Any, **kwargs: Any) -> ComponentProtocol: ...

    def test_flows_protocol(self, comp_fixture: ComponentProtocol) -> None:
        assert isinstance(comp_fixture, ComponentProtocol)

    def test_comp_protocol_attrs(self, comp_fixture: ComponentProtocol) -> None:
        fp = FunctionParser()
        comp_vars = fp.comp_vars(comp_fixture)  # type: ignore
        assert comp_vars
        assert fp.get_func_params(comp_vars)

    def test_comp_protocol_e2e(self, comp_fixture: ComponentProtocol) -> None:
        fp = FunctionParser()
        fp.create_function(comp_fixture)



import dataclasses
from typing import List

from ml_orchestrator.comp_protocol.comp_protocol import ComponentProtocol
from ml_orchestrator.comp_protocol.func_parser import FunctionParser
from ml_orchestrator.env_params import EnvironmentParams
from ml_orchestrator.meta_comp import MetaComponent


@dataclasses.dataclass
class ComponentParser(FunctionParser):
    add_imports: List[str] = dataclasses.field(default_factory=lambda: [])
    only_function: bool = False

    @staticmethod
    def create_decorator(env: EnvironmentParams) -> str:
        dec_vars = env.comp_vars()
        prams = ComponentParser.get_func_params(dec_vars, with_typing=False)
        override_params = ComponentParser._get_decorator_override_params(prams)
        dec_scope = "(\n\t" + ",\n\t".join(override_params) + "\n)"
        dec_scope = ComponentParser.convert_to_format_str(dec_scope)
        func_definition = f"@component{dec_scope}"
        return func_definition

    @staticmethod
    def convert_to_format_str(text: str) -> str:
        new_text = text.replace(", '", ", f'").replace("['", "[f'")
        new_text = new_text.replace(', "', ', f"').replace('["', '[f"')
        return new_text

    @staticmethod
    def _get_decorator_override_params(prams: List[str]) -> List[str]:
        return [p for p in prams if "None" not in p]

    def create_kfp_str(self, component: MetaComponent) -> str:  # type: ignore
        function_str = super().create_kfp_str(component)  # type: ignore
        if self.only_function:
            return function_str
        decorator_str = self.create_decorator(component.env)
        decorator_str = decorator_str.replace(" = ", "=")
        kfp_component_str = f"{decorator_str}\n{function_str}"
        kfp_component_str = kfp_component_str.replace("\t", "    ")
        return kfp_component_str + "\n"

    def parse_components_to_file(self, components: List[ComponentProtocol], filename: str) -> None:
        kfp_str = self.create_kfp_file_str(components)
        self.write_to_file(filename, kfp_str)

    def write_to_file(self, filename: str, file_content: str) -> None:
        for imp in self.add_imports:
            file_content = f"{imp}\n{file_content}"
        return super().write_to_file(filename, file_content)



from typing import Any, Protocol, Union, runtime_checkable

ComponentOutput = Union[None, str, int, float, bool, list, dict]


@runtime_checkable
class ComponentProtocol(Protocol):
    def execute(self) -> ComponentOutput: ...

    def __dataclass_fields__(self) -> Any: ...



import abc
import dataclasses
from typing import Any, Dict

from ml_orchestrator.env_params import EnvironmentParams


@dataclasses.dataclass
class MetaComponent(abc.ABC):
    @abc.abstractmethod
    def execute(
        self,
    ) -> None: ...

    @classmethod
    def kfp_func_name(cls) -> str:
        return cls.__name__.lower()

    def comp_vars(self) -> Dict[dataclasses.Field, Any]:
        fields = dataclasses.fields(self.__class__)
        ins_vars = dict()
        for field in fields:
            ins_vars[field] = getattr(self, field.name)

        return ins_vars

    @property
    def env(self) -> EnvironmentParams:
        return EnvironmentParams()



import dataclasses
from typing import Any, Dict, List, Tuple, Type, Union

from ml_orchestrator.utils.field_utils import get_param_meta_data, get_param_meta_data_str

from .comp_protocol import ComponentProtocol

IMPORT_COMPOUND = "from kfp.dsl import *\nfrom typing import *\nfrom importlib.metadata import version\n"


@dataclasses.dataclass
class FunctionParser:
    @classmethod
    def get_class_def(cls, component: Union[type, ComponentProtocol]) -> type:
        is_class_def = isinstance(component, type)
        com_class = component if is_class_def else component.__class__
        return com_class  # type: ignore

    def create_function(self, component: ComponentProtocol) -> str:
        comp_class = self.get_class_def(component)
        comp_init, func_definition = self.get_function_parts(comp_class)
        import_compound = f"from {comp_class.__module__} import {comp_class.__name__}"
        comp_run = self.comp_execute(comp_class)
        str_func_body = "\n\t".join([import_compound, comp_init, comp_run])
        str_func = func_definition + "\n\t" + str_func_body
        return str_func

    def get_function_parts(self, comp_class: Type[ComponentProtocol]) -> Tuple[str, str]:
        component_variables = self.comp_vars(comp_class)
        kfp_func_name = comp_class.__name__.lower()
        func_scope = "(\n\t" + ",\n\t".join(self.get_func_params(component_variables)) + "\n)"
        comp_scope = "(\n\t\t" + ",\n\t\t".join(self.get_comp_params(component_variables)) + "\n\t)"
        func_definition = f"def {kfp_func_name}{func_scope}:"
        comp_init = f"comp = {comp_class.__name__}{comp_scope}"
        return comp_init, func_definition

    @classmethod
    def comp_execute(cls, comp: Type[ComponentProtocol]) -> str:
        execute_return_type = comp.execute.__annotations__.get("return")
        comp_run = "comp.execute()"
        if execute_return_type is not None:
            comp_run = "return comp.execute()"
        return comp_run

    @staticmethod
    def get_func_params(comp_vars: Dict[dataclasses.Field, Any], with_typing: bool = True) -> List[str]:
        return [
            get_param_meta_data_str(*get_param_meta_data(k, v), with_typing=with_typing) for k, v in comp_vars.items()
        ]

    @classmethod
    def comp_vars(cls, component: Type[ComponentProtocol]) -> Dict[dataclasses.Field, Any]:
        fields = dataclasses.fields(component)  # type: ignore
        ins_vars = dict()
        for field in fields:
            ins_vars[field] = getattr(component, field.name)

        return ins_vars

    @staticmethod
    def get_comp_params(comp_vars: Dict[dataclasses.Field, Any]) -> List[str]:
        return [f"{k.name}={k.name}" for k, v in comp_vars.items()]

    def create_kfp_str(self, component: ComponentProtocol) -> str:
        function_str = self.create_function(component)
        return function_str

    def create_kfp_file_str(
        self,
        components: List[ComponentProtocol],
    ) -> str:
        kfp_str = ""
        for component in components:
            kfp_str += self.create_kfp_str(component)

            kfp_str += "\n\n"

        file_content = f"{IMPORT_COMPOUND}\n\n\n{kfp_str}"
        return file_content

    def write_to_file(self, filename: str, file_content: str) -> None:
        file_content = f"# flake8: noqa: F403, F405, B006\n{file_content}"
        with open(filename, "w", encoding="utf-8") as f:
            f.write(file_content)



from importlib.metadata import version

from .comp_parser import ComponentParser
from .comp_protocol.func_parser import FunctionParser
from .env_params import EnvironmentParams
from .meta_comp import MetaComponent

__version__ = version("ml-orchestrator")
__all__ = ["ComponentParser", "FunctionParser", "MetaComponent", "EnvironmentParams"]
